# $Id: Courier.pm,v 1.5 2005/01/13 00:14:32 braeucup Exp $

#
# MTA module for use as the Courier-MTA perlfilter
# By: John Laur <johnl@blurbco.com>
# Based on AMAVIS::MTA::Exim
#

package AMAVIS::MTA::Courier;
use strict;
use vars qw($VERSION);
$VERSION='0.1';

use AMAVIS;
use AMAVIS::Logging;
use IO::File;
use File::Path;

use File::Copy;
use Sys::Hostname;

use vars qw(
            $cfg_sendmail_binary
	    $cfg_sendmail_args
	    $cfg_x_header
	    $cfg_x_header_tag
	    $cfg_x_header_line
	    $perlfilter_socket
	    $return_code
	   );

# Init and open the socket to communicate with Courier
sub init {
  my $self = shift;
  my $args = shift;

  # Get the configuration for sending notificatin messages
  $cfg_sendmail_binary = $AMAVIS::cfg->val('Courier', 'sendmail');
  $cfg_sendmail_args = $AMAVIS::cfg->val('Courier', 'args');
  if ($cfg_sendmail_binary eq '') {
    writelog($args,LOG_CRIT, "Location of courier sendmail binary not specified");
    return 0;
  }
  if (! -x $cfg_sendmail_binary) {
    writelog($args,LOG_CRIT, "$cfg_sendmail_binary not executable");
    return 0;
  }
  
  # Get the configuration for adding a custom header to the message
  if (($AMAVIS::cfg->val('global', 'x-header') || '') eq 'true') {
    $cfg_x_header = 1
  };
  $cfg_x_header_tag = $AMAVIS::cfg->val('global', 'x-header-tag');
  $cfg_x_header_line = $AMAVIS::cfg->val('global', 'x-header-line');
  writelog($args,LOG_DEBUG,__PACKAGE__." initialized.");
  
  # The fd of our perlfilter socket is passed to us on STDIN
  my $socket_fd = shift @ARGV;
  $perlfilter_socket = IO::File->new("+<&$socket_fd");
  unless (defined $perlfilter_socket) {
    writelog($args,LOG_ERR, __PACKAGE__.": Could not open socket");
    return 0;
  }

  # Return successfully
  return 1;
}

# Return the message result code to courier (and on to the client) and close socket
sub cleanup {
  my $self = shift;
  my $args = shift;

  if ($return_code eq '') {
    $return_code = '550 Message rejected due to unknown error during virus scan.';
    writelog($args,LOG_ERR,__PACKAGE__.": Unknown Error During Scan. Dumping Message");
  }

  #Print the return message back to the socket
  print $perlfilter_socket $return_code . "\n";
  
  #Close the socket
  close($perlfilter_socket);

  return 1;
}

# Create temp dir and write mail
sub get_directory {
  my $self = shift;
  my $args = shift;

  # Create temp directory. Try creating $prefix[date]-[pid] 
  # If unsuccessful after 10 tries, abort
  my $prefix = "$AMAVIS::cfg_unpack_dir/amavis-unpack-";
  my $i = 0;
  my $message_id;
  while (1) {
    $message_id = sprintf("%.8x-%.4x",time,$$);
    unless (defined mkpath ($prefix.$message_id, 0, 0775)) {
      if (++$i > 10) {
	return 0;
      }
      else {
	next;
      }
    }
    last;
  }
  $$args{'message_id'}=$message_id;
  $$args{'directory'} = $directory;
  my $directory = $prefix.$message_id;
  mkdir "$directory/parts", 0777;

  # Open message file that is later going to be disected and scanned
  my $output = IO::File->new("+>$directory/email.txt");

  # Get the filenames for the message and control files from the socket
  my @controlfiles = ();
  while (<$perlfilter_socket>) {
    chomp;
    last unless $_;
    push @controlfiles, $_;
  }
  $$args{'messagefile'} = shift @controlfiles;
  
  writelog($args,LOG_DEBUG,__PACKAGE__.": Message file is $$args{'messagefile'}");
  writelog($args,LOG_DEBUG,__PACKAGE__.": Control files are "
	   .join(', ',@controlfiles));

  # Read message from file
  my $message = IO::File->new("<$$args{'messagefile'}");
  my $headers=1;
  $$args{'headers'}='';
  while (<$message>) {
    $headers=0 if /^ *$/;
    $$args{'headers'}.=$_ if $headers==1;
    #TODO Skip this whole message if we have our 'secret' header that lets us know
    #we've already scanned it. Courier doesnt allow for certain messages to skip
    #filters very well yet and we don't want to create any mail loops.
    print $output $_;
  }
  close($message);

  #Parse control files to get senders and recipients
  my %control = ();
  foreach my $controlfile (@controlfiles) {
    my $control_fh = new IO::File $controlfile;
    while (<$control_fh>) {
      chomp;
      $$args{'sender'} = $1 if /^s(.*)$/;
      push @{$$args{'recipients'}}, $1 if /^r(.*)$/;
    }
    close($control_fh);
  }
  $$args{'sender'} = "<>" if (!$$args{'sender'});
  
  writelog($args,LOG_DEBUG,__PACKAGE__.": Sender is $$args{'sender'}");
  writelog($args,LOG_DEBUG,__PACKAGE__.": Recipients are "
	   .join(', ',@{$$args{'recipients'}}));

  # Message file has been written, reset file pointer and put it into
  # the record.
  $output->seek(0,0);
  $$args{'filehandle'} = $output;

  # Return successfully
  return 1;
}

# Accept the message for delivery and rewrite if needed.
sub accept_message {
  my $self = shift;
  my $args = shift;
  writelog($args,LOG_INFO, __PACKAGE__.": Accepting message");

  #TODO Courier filters cannot rewrite MIME messages, we'd like
  #to be able to add a header or something to this message but
  #we can't.(yet). Sorry
  
  ##Rewrite the message in the queue if we need to add a header
  #if ($cfg_x_header) {
  #  my $queuefile = IO::File->new("+>$$args{'messagefile'}");
  #  print $queuefile "$cfg_x_header_tag: $cfg_x_header_line\n";
  #  while (my $line=$$args{'filehandle'}->getline()) {
  #    print $queuefile $line;
  #  }
  #  close($queuefile);
  #}

  $$args{'status'} = 'accept';
  
  $return_code = '200 Ok';

  # Return successfully
  return 1;
}

# Just "forget" the mail.
sub drop_message {
  my $self = shift;
  my $args = shift;
  # Doing nothing should suffice
  writelog($args,LOG_WARNING, __PACKAGE__.": Dropping message");

  $return_code = '550 Message rejected by virus scanner due to possible virus.';

  # Return successfully
  return 1;
}

# Courier can't really freeze, so we reject and quarantine the mail. 
sub freeze_message {
  my $self = shift;
  my $args = shift;
  writelog($args,LOG_WARNING, __PACKAGE__.": Freezing message");

  # Instead of freezing, drop the message, but store a copy in
  # $cfg_problem_dir -- analogous to AMAVIS::quarantine_message()
  if (AMAVIS->quarantine_problem_message($args)) {
    
    $return_code = '550 Message quarantined by virus scanner due to possible virus.'
         .' Your message may still be delivered after further inspection.';

    # Return successfully
    return 1;
  }
  else {
    writelog($args,LOG_ERR,__PACKAGE__.": Could not freeze message");

    $return_code = '550 Message rejected by virus scanner due to possible virus.'
         .' Quarnatine of message failed. Message will not be delivered.';

    return 0;
  }
}

# Called from Notify::*.pm, i.e. for sending warining messages
sub send_message {
  my $self = shift;
  my $args = shift;
  my $message = shift;
  my $sender = shift;
  my @recipients = @_;
  writelog($args,LOG_DEBUG, __PACKAGE__.": Sending mail from $sender to ".
	   join(', ',@recipients));

  my @cfg_sendmail_args;

  push @cfg_sendmail_args, split(/\s+/,$cfg_sendmail_args);

  writelog($args,LOG_DEBUG, __PACKAGE__.": Running $cfg_sendmail_binary ".
    join(' ',@cfg_sendmail_args));

  open(MAIL, "|-") || exec($cfg_sendmail_binary, @cfg_sendmail_args);
  print MAIL $message;
  close(MAIL);

  if ($? != 0) {
    writelog($args,LOG_ERR,
      __PACKAGE__.": $cfg_sendmail_binary @cfg_sendmail_args exited with ".($?>>8));
  }

  # Return successfully
  return 1;
}

1;
